from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import mean_squared_error
from sklearn.base import BaseEstimator
from abc import abstractmethod, ABCMeta
import numpy as np
from tqdm import tqdm


class SimilarityForestRegressor(BaseEstimator, metaclass=ABCMeta):
    """
    https://www.ncbi.nlm.nih.gov/pmc/articles/PMC2822360/
    """

    def __init__(self, models_parameters, score_metric=mean_squared_error):
        self._models_parameters = models_parameters
        self._estimator = RandomForestRegressor(**self._models_parameters.hyperparameters,
            random_state=self._models_parameters.seed, n_jobs=-1)
        self._extracted_forest_size = self._models_parameters.extracted_forest_size
        self._score_metric = score_metric
        self._selected_trees = list()

    @property
    def models_parameters(self):
        return self._models_parameters

    @property
    def selected_trees(self):
        return self._selected_trees

    def fit(self, X_train, y_train, X_val, y_val):
        self._estimator.fit(X_train, y_train)

        y_val_pred = self._estimator.predict(X_val)
        forest_pred = self._score_metric(y_val, y_val_pred)
        forest = self._estimator.estimators_
        tree_list = list(self._estimator.estimators_)

        val_scores = list()
        with tqdm(tree_list) as tree_pred_bar:
            tree_pred_bar.set_description('[Initial tree predictions]')
            for tree in tree_pred_bar:
                val_scores.append(tree.predict(X_val))
            tree_pred_bar.update(1)

        with tqdm(range(self._extracted_forest_size), disable=True) as pruning_forest_bar:
            pruning_forest_bar.set_description(f'[Pruning forest s={self._extracted_forest_size}]')
            for i in pruning_forest_bar:
                best_similarity = 100000
                found_index = 0
                with tqdm(range(len(tree_list)), disable=True) as tree_list_bar:
                    tree_list_bar.set_description(f'[Tree selection s={self._extracted_forest_size} #{i}]')
                    for j in tree_list_bar:
                        lonely_tree = tree_list[j]
                        del tree_list[j]
                        val_mean = np.mean(np.asarray(val_scores), axis=0)
                        val_score = self._score_metric(val_mean, y_val)
                        temp_similarity = abs(forest_pred - val_score)
                        if (temp_similarity < best_similarity):
                            found_index = j
                            best_similarity = temp_similarity
                        tree_list.insert(j, lonely_tree)
                        val_scores.insert(j, lonely_tree.predict(X_val))
                        tree_list_bar.update(1)
                self._selected_trees.append(tree_list[found_index])
                del tree_list[found_index]
                del val_scores[found_index]
                pruning_forest_bar.update(1)

        self._selected_trees = set(self._selected_trees)
        pruned_forest = list(set(forest) - self._selected_trees)
        self._estimator.estimators_ = pruned_forest

    def score(self, X, y):
        test_list = list()
        for mod in self._estimator.estimators_:
            test_pred = mod.predict(X)
            test_list.append(test_pred)
        test_list = np.array(test_list)
        test_mean = np.mean(test_list, axis=0)
        score = self._score_metric(test_mean, y)
        return score

    def predict_base_estimator(self, X):
        return self._estimator.predict(X)
